#include <iostream>
#include <windows.h>
#include <dshow.h>
#include <vector>
#include <string>
#include <thread>

#include "opencv2\core\core.hpp"
#include "opencv2\highgui\highgui.hpp"

#pragma comment(lib, "strmiids")

using namespace cv;
using namespace std;

HRESULT EnumerateDevices(REFGUID category, IEnumMoniker **ppEnum);
string ConvertWCSToMBS(const wchar_t* pstr, long wslen);
string ConvertBSTRToMBS(BSTR bstr);
void ListDevices(IEnumMoniker * pEnum, vector<string> * resultList);
void getDeviceListB(vector<string> * resultList);
vector<string> getDeviceList();

int main()
{
	// Get device list
	vector<string> deviceList = getDeviceList();
	const int numDevices = deviceList.size();

	// Check if devices were found
	if (numDevices == 0)
	{
		cout << "No compatible devices were found. Press Enter to exit." << endl;
		getchar();
		return 0;
	}

	// List the devices with device index
	cout << "The following devices were found: " << endl;
	int dispIndex = 0;
	for (vector<string>::iterator it = deviceList.begin(); it != deviceList.end(); ++it)
		cout << dispIndex++ << " - " << *it << endl;
	cout << endl;

	// Ask user to select a device
	int a = -1;
	string mystr;
	while (true) {
		cout << "Select a device by number and press Enter: ";
		getline(cin, mystr);
		cout << endl;

		try {
			a = stoi(mystr);
		} catch (...) {
			a = -1;
		}

		if (a >= 0 && a < numDevices)
			break;
		else
			cout << "Invalid number" << endl;
	}

	// Try to open the device
	VideoCapture cap(a);
	if (cap.isOpened())
	{
		cout << "Device succesfully opened!" << endl;
		cout << "Attempting to read image data..." << endl;
		cout << endl;

		// Try to read and display image data
		const int numRequiredFrames = 60;
		namedWindow("Preview");
		Mat frame;
		int successes = 0;
		int failures = 0;
		while (successes < numRequiredFrames)
		{
			try {
				cap >> frame;

				if (frame.size().height > 0 && frame.size().width > 0)
				{
					imshow("Preview", frame);
					waitKey(30);
					successes++;
				}
				else
				{
					break;
				}
			} catch (...) {
				break;
			}
		}
		destroyAllWindows();
		cap.release();

		// If image data could be read, it is compatible, show user the result
		const bool compatible = successes >= numRequiredFrames;
		if (compatible)
		{
			cout << "Your webcam is compatible with OpenCV." << endl;
		}
		else
		{
			cout << "Your webcam can not be read by OpenCV." << endl;
			cout << "If possible, try the following: " << endl;
			cout << "- install/update latest drivers" << endl;
			cout << "- disable webcam splitting software" << endl;
			cout << "- use a different webcam" << endl;
		}

		cout << "Press Enter to exit." << endl;
		getchar();
		return 0;
	}
	else
	{
		cout << "Device could not be opened. Press Enter to exit." << endl;
		getchar();
		return 0;
	}
}


HRESULT EnumerateDevices(REFGUID category, IEnumMoniker **ppEnum)
{
    // Create the System Device Enumerator.
    ICreateDevEnum *pDevEnum;
    HRESULT hr = CoCreateInstance(CLSID_SystemDeviceEnum, NULL, CLSCTX_INPROC_SERVER, IID_PPV_ARGS(&pDevEnum));

    if (SUCCEEDED(hr))
    {
        // Create an enumerator for the category.
        hr = pDevEnum->CreateClassEnumerator(category, ppEnum, 0);
        if (hr == S_FALSE)
        {
            hr = VFW_E_NOT_FOUND;  // The category is empty. Treat as an error.
        }
        pDevEnum->Release();
    }
    return hr;
}

string ConvertWCSToMBS(const wchar_t* pstr, long wslen)
{
    int len = ::WideCharToMultiByte(CP_ACP, 0, pstr, wslen, NULL, 0, NULL, NULL);
    string dblstr(len, '\0');
    len = ::WideCharToMultiByte(CP_ACP, 0, pstr, wslen, &dblstr[0], len, NULL, NULL);
    return dblstr;
}

string ConvertBSTRToMBS(BSTR bstr)
{
    int wslen = ::SysStringLen(bstr);
    return ConvertWCSToMBS((wchar_t*)bstr, wslen);
}

void ListDevices(IEnumMoniker * pEnum, vector<string> * resultList)
{
    IMoniker *pMoniker = NULL;
	int slot = 0;

    while (pEnum->Next(1, &pMoniker, NULL) == S_OK)
	{
		IPropertyBag *pPropBag;
		HRESULT hr = pMoniker->BindToStorage(0, 0, IID_PPV_ARGS(&pPropBag));
		if (FAILED(hr))
		{
			string defaultName = "Device ";
			defaultName.append(to_string(slot++));
			resultList->push_back(defaultName);
			pMoniker->Release();
			continue;
		}

        VARIANT var;
        VariantInit(&var);
        
		hr = pPropBag->Read(L"FriendlyName", &var, 0);
		if (SUCCEEDED(hr))
		{
			string friendlyName = ConvertBSTRToMBS(var.bstrVal);
			resultList->push_back(friendlyName);
			VariantClear(&var);
		}
		else
		{
			string defaultName = "Device ";
			defaultName.append(to_string(slot));
			resultList->push_back(defaultName);
		}
		slot++;
		pPropBag->Release();
		pMoniker->Release();
	}
}

void getDeviceListB(vector<string> * resultList)
{
	HRESULT hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);
	if (SUCCEEDED(hr))
	{
	    IEnumMoniker *pEnum;
	
	    HRESULT hr = EnumerateDevices(CLSID_VideoInputDeviceCategory, &pEnum);
	    if (SUCCEEDED(hr))
	    {
	        ListDevices(pEnum, resultList);
	        pEnum->Release();
	    }
	    CoUninitialize();
	}
}

vector<string> getDeviceList()
{
	// New thread because thread mode must be COINIT_MULTITHREADED to request device list
	vector<string> resultList;
	thread thread(getDeviceListB, &resultList);
	thread.join();
	return resultList;
}